import pickle

import pandas as pd 
import numpy as np

from ..metrics.coco_metrics import COCOMetrics

class PlotViolin:
    def __init__(self, class_mappings, ground_truth_dict=None, prediction_dict=None, artifacts_path=None, study_list=None):
        if ground_truth_dict:
            self.ground_truth_dict = ground_truth_dict
            self.prediction_dict = prediction_dict
            self.class_mappings = class_mappings
            self.coco_metrics = COCOMetrics(class_mappings)
            self.artifacts_path = artifacts_path
            self.study_list = study_list

    def violin_graph(self):

        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(self.ground_truth_dict, self.prediction_dict)

        # TO DO: add imagewise APD in its function
        # Add in coco_metrics.py plots as well
        iou_series = pd.DataFrame.from_dict(artifacts["iou"]["imagewise_iou"], orient='index',
                                            columns=['imagewise_iou']).sort_values(by='imagewise_iou', ascending=True)
        fwiou_series = pd.DataFrame.from_dict(artifacts["fwiou"]["imagewise_fwiou"], orient='index',
                                            columns=['imagewise_fwiou']).sort_values(by='imagewise_fwiou', ascending=True)
        acc_series = pd.DataFrame.from_dict(artifacts["pAccs"]["imagewise_acc"], orient='index',
                                            columns=['imagewise_acc']).sort_values(by='imagewise_acc', ascending=True)
        

        dice_series = self.create_miseval_imagewise_df(artifacts, 'DSC')
        spec_series = self.create_miseval_imagewise_df(artifacts, 'Specificity')
        sens_series = self.create_miseval_imagewise_df(artifacts, 'Sensitivity')
        kapp_series = self.create_miseval_imagewise_df(artifacts, 'Kapp')
        auc_series = self.create_miseval_imagewise_df(artifacts, 'AUC')

        payload_dict = {"type": "violin",
                        "data": { "IoU": iou_series['imagewise_iou'].tolist(),
                                "FWIoU": fwiou_series['imagewise_fwiou'].tolist(),
                                "Acc": acc_series['imagewise_acc'].tolist(),
                                "Dice": dice_series['DSC'].tolist(),
                                "Spec": spec_series['Specificity'].tolist(),
                                "Sens": sens_series['Sensitivity'].tolist(),
                                "Kapp": kapp_series['Kapp'].tolist(),
                                "AUC": auc_series['AUC'].tolist()}}
                                                    
        return payload_dict



    def create_miseval_imagewise_df(self, artifacts, metric_name):
        metric_values = [(img_id, metrics[metric_name]) for img_id, metrics in artifacts['misevals']['imagewise_metrics'].items()]
        df = pd.DataFrame(metric_values, columns=['Image ID', metric_name]).sort_values(metric_name)
        return df

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)


from itertools import chain

import numpy as np
import pandas as pd
import random

from ..metrics.average_precision import AveragePrecision
from gesund_val_library.utils.validation_data_utils import ValidationUtils, Statistics

class PlotConfidenceGraphs:
    def __init__(self, class_mappings, coco_, meta_data_dict=None):
        self.class_mappings = class_mappings
        self.coco_ = coco_
        if bool(meta_data_dict):
            self.is_meta_exists = True
            self.validation_utils = ValidationUtils(meta_data_dict)

    def _plot_confidence_histogram_scatter_distribution(
        self, predicted_class=None, n_samples=300
    ):

        average_precision = AveragePrecision(
            class_mappings=self.class_mappings,
            coco_=self.coco_,
        )

        threshold = 0.1
        conf_points = average_precision.calculate_confidence_distribution(
            threshold, predicted_class=predicted_class
        )
        conf_points_list = []
        for neuron, value in conf_points.items():
            conf_points_list.append(
                {
                    "image_id": neuron,
                    "y": value,
                    "x": random.uniform(0, 1),
                    "labels": "TP",
                }
            )

        y_values = [d["y"] for d in conf_points_list]

        # Plot histogram
        histogram = Statistics.calculate_histogram(y_values, min_=0, max_=1, n_bins=10)

        payload_dict = {
            "type": "mixed",
            "data": {"points": conf_points_list, "histogram": histogram},
        }

        return payload_dict



import argparse
import json
import os
import bson
from pathlib import Path

def read_json(file_path):
    """
    Read a JSON file and return the data.
    
    Args:
    file_path (str): Path to the JSON file.
    
    Returns:
    dict: The JSON data loaded into a Python dictionary.
    """
    with open(file_path, 'r') as file:
        data = json.load(file)
    return data

def save_plot_metrics_as_json(overall_metrics, output_dir):
    """Saves each plot's metrics as individual JSON files."""
    # Create output directory if it doesn't exist
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    
    # Iterate over the overall metrics and save each as a separate JSON
    for plot_name, metrics in overall_metrics.items():
        output_file = os.path.join(output_dir, f"{plot_name}.json")
        try:
            with open(output_file, 'w') as f:
                json.dump(metrics, f, indent=4)
        except Exception as e:
            print(f"Could not save metrics for {plot_name} because: {e}")
            
            
def format_metrics(metrics):
    """ Format and print the metrics in a readable way """
    print("\nOverall Highlighted Metrics:\n" + "-"*40)
    for metric, values in metrics['plot_highlighted_overall_metrics']['data'].items():
        print(f"{metric}:")
        for key, value in values.items():
            if isinstance(value, list):  # If it's a confidence interval
                value_str = f"{value[0]:.4f} to {value[1]:.4f}"
            else:
                value_str = f"{value:.4f}"
            print(f"    {key}: {value_str}")
        print("-"*40)
    print("All Graphs and Plots Metrics saved in JSONs.\n" + "-"*40)



#!/usr/bin/env python
import json
from gesund_val_library.utils.io_utils import read_json, validate_json_format
from gesund_val_library.metrics.classification.create_validation import ValidationCreation

def check_json_fields(data, required_fields):
    """
    Check if each item in the JSON data contains all required fields.
    
    Args:
    data (dict): The JSON data to check.
    required_fields (list): List of fields that should be present in each item.
    
    Raises:
    AssertionError: If any field is missing in the data.
    """
    for item_id, item in data.items():
        missing_fields = [field for field in required_fields if field not in item]
        if missing_fields:
            raise AssertionError(f"Item ID {item_id} is missing fields: {missing_fields}")

def test_create_validation_collection_data():
    batch_job_id = "test_batch_job_id"
    successful_batch_data_file = 'test_data/test_predictions_classification.json'
    annotation_data_file = 'test_data/test_annotations_classification.json'
    class_mappings_file = 'test_data/class_mappings.json'
    
    # Read data from JSON files
    successful_batch_data = read_json(successful_batch_data_file)
    annotation_data = read_json(annotation_data_file)
    class_mappings = read_json(class_mappings_file)

    # Define required fields for each type of data
    successful_batch_fields = ['image_id', 'confidence', 'logits', 'prediction_class', 'loss']
    annotation_fields = ['image_id', 'annotation']

    # Validate the format of the JSON data
    validate_json_format(successful_batch_data, successful_batch_fields)
    validate_json_format(annotation_data, annotation_fields)
    # Check if all required fields are present in the JSON files
    check_json_fields(successful_batch_data, successful_batch_fields)
    check_json_fields(annotation_data, annotation_fields)
    
    # Create ValidationCreation instance and process data
    validation = ValidationCreation(batch_job_id)
    validation_data = validation.create_validation_collection_data(successful_batch_data, annotation_data)
    metrics = validation.load(validation_data, class_mappings)

    # Print or assert results for further validation
    print(metrics)
    # Example assertions (customize based on your requirements)
    assert isinstance(validation_data, list)
    assert len(validation_data) > 0
    assert all("image_id" in item for item in validation_data)
    assert all("confidence" in item for item in validation_data)
    assert all("logits" in item for item in validation_data)
    assert all("prediction_class" in item for item in validation_data)
    assert all("ground_truth" in item for item in validation_data)
    assert all("created_timestamp" in item for item in validation_data)
    assert all("loss" in item for item in validation_data)

if __name__ == "__main__":
    test_create_validation_collection_data()




import random
import pickle 

import pandas as pd 
import numpy as np

from ..metrics.coco_metrics import COCOMetrics
from gesund_val_library.utils.validation_data_utils import ValidationUtils, Statistics

class PlotIoUDistribution:
    def __init__(self, class_mappings, ground_truth_dict=None, prediction_dict=None, artifacts_path=None, study_list=None):
        if ground_truth_dict:
            self.ground_truth_dict = ground_truth_dict
            self.prediction_dict = prediction_dict
            self.class_mappings = class_mappings
            self.coco_metrics = COCOMetrics(class_mappings)
            self.artifacts_path = artifacts_path
            self.study_list = study_list

    def iou_distribution(self, n_samples=300):
        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(self.ground_truth_dict, self.prediction_dict)

        iou_series = pd.Series(artifacts["iou"]["imagewise_iou"]).sort_values()
        iou_series = iou_series.sample(min(n_samples,iou_series.size)).to_dict()

        points = []

        for image_id in iou_series:
            iou_point = {
                    "image_id": image_id,
                    "x": random.uniform(0,1),

                    "y": float(iou_series[image_id]),
                }
            points.append(iou_point)

        # Plot Histogram
        histogram = Statistics.calculate_histogram(pd.Series(iou_series.values()), min_=0, max_=1, n_bins=10
        )
            
        payload_dict =payload_dict = {"type": "mixed","data": {"points": points, "histogram": histogram}}

        return payload_dict

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



import pandas as pd
import numpy as np
import pickle
import math
import os

from ..metrics import COCOMetrics

class PlotTopMisses:
    def __init__(self, ground_truth_dict, prediction_dict, class_mappings, artifacts_path, study_list=None):
        self.ground_truth_dict = ground_truth_dict
        self.prediction_dict = prediction_dict
        self.class_mappings=class_mappings
        self.artifacts_path = artifacts_path
        self.study_list = study_list

    def top_misses(self, metric, sort_by, top_k=150):
        try:
            artifacts = self._load_pickle(self.artifacts_path)
        except:
            coco_metrics = COCOMetrics(self.class_mappings)
            artifacts = coco_metrics.create_artifacts(self.ground_truth_dict, self.prediction_dict)

        additional_metrics = artifacts.keys()
        top_misses_data = []
        i = 1  # FIX.

        if metric in additional_metrics:
            for image_id, value in pd.DataFrame(
                artifacts['{}'.format(metric)]).sort_values('imagewise_{}'.format(metric), ascending=(sort_by == "Ascending")
                                                            ).head(top_k).to_dict()['imagewise_{}'.format(metric)].items():

                top_loss_single_image = {
                        "image_id": image_id,
                        "rank": i,
                        metric: value,
                }
                top_misses_data.append(top_loss_single_image)
                i += 1
        else:
            for image_id, value in pd.DataFrame(artifacts['misevals']['imagewise_metrics']).T.sort_values(metric,ascending=(sort_by == "Ascending")).head(top_k)[metric].to_dict().items():
                top_loss_single_image = {
                        "image_id": image_id,
                        "rank": i,
                        metric: value,
                    }
                top_misses_data.append(top_loss_single_image)
                i += 1

        payload_dict = {"type": "image", "data": top_misses_data}
        return payload_dict

    def _load_pickle(self, file_path):
        with open(file_path, "rb") as f:
            return pickle.load(f)



import time
import pandas as pd
import numpy as np

from .plots.plot_driver import ClassificationPlotDriver

class ValidationCreation:
    def __init__(self, batch_job_id, filter_field="image_url", generate_metrics=True):
        self.batch_job_id = batch_job_id
        self.filter_field = filter_field
        self.generate_metrics = generate_metrics

    def create_validation_collection_data(self, successful_batch_data, annotation_data, format=None):
        """
        Create validation collection data from JSON files.
        
        Args:
        successful_batch_data_file (str): Path to the JSON file with prediction data.
        annotation_data_file (str): Path to the JSON file with annotation data.
        
        Returns:
        list: List of dictionaries with validation collection data.
        """

        validation_collection_data = []

        for item_id in successful_batch_data:
            if item_id not in annotation_data:
                raise ValueError(f"Annotation data for item ID {item_id} is missing")

            batch_item = successful_batch_data[item_id]
            annotation_item = annotation_data[item_id]

            information_dict = {
                "batch_job_id": self.batch_job_id,
                "image_id": batch_item["image_id"],
                "confidence": batch_item["confidence"],
                "logits": batch_item["logits"],
                "prediction_class": batch_item["prediction_class"],
                "loss": batch_item["loss"],
                "ground_truth": annotation_item["annotation"][0]["label"],
                "meta_data": {},  # Placeholder for meta data if needed
                "created_timestamp": time.time()
            }

            validation_collection_data.append(information_dict)            

        return validation_collection_data
    
    def load(self, validation_collection_data, class_mappings):

        generate_metrics = True
        plotting_data = self._load_plotting_data(validation_collection_data)

        # Create per image variables
        ground_truth_dict = plotting_data["per_image"]["ground_truth"]
        meta_data_dict = plotting_data["per_image"]["meta_data"]
        logits_dict = plotting_data["per_image"]["logits"]
        if generate_metrics:
            loss_dict = plotting_data["per_image"]["loss"]

        # Create validation variables
        true = pd.DataFrame(ground_truth_dict, index=[0]).loc[0].astype(int)
        class_mappings = class_mappings
        pred = pd.DataFrame(logits_dict)
        meta = pd.DataFrame(meta_data_dict).T

        loss = None
        if generate_metrics:
            loss = pd.DataFrame(loss_dict, index=[0])
        sample_size = len(true)
        class_order = list(range(len(class_mappings.keys())))
        true.name = "true"
        pred.name = "pred"
        # Check, add breakpoint
        if np.shape(pred.shape)[0] == 1:
            pred_categorical = pred
            pred_categorical.name = "pred_categorical"
            pred_categorical = pred_categorical.astype(int)
        else:
            pred_logits = pred
            pred_categorical = pred.idxmax()
            pred_categorical.name = "pred_categorical"
        meta_pred_true = pd.concat([meta, pred_categorical, true], axis=1)

        self.plot_driver = ClassificationPlotDriver(
            true=true,
            pred=pred,
            meta=meta,
            pred_categorical=pred_categorical,
            pred_logits=pred_logits,
            meta_pred_true=meta_pred_true,
            class_mappings=class_mappings,
            loss=loss,
        )
        
        overall_metrics = self.plot_driver._calling_all_plots()
        
        return overall_metrics
        
    def _load_plotting_data(
        self, validation_collection_data
    ):
        plotting_data = dict()
        plotting_data["per_image"] = self._craft_per_image_plotting_data(
            validation_collection_data
        )
        return plotting_data

    def _craft_per_image_plotting_data(
        self, validation_collection_data
    ):
        data = dict()
        validation_df = pd.DataFrame(validation_collection_data)
        # Ground truth dict
        ground_truth_dict = validation_df[["image_id", "ground_truth"]].values
        ground_truth_dict = dict(zip(ground_truth_dict[:, 0], ground_truth_dict[:, 1]))
        # Loss dict
        loss_dict = (
            validation_df[["image_id", "loss"]]
            .set_index("image_id")
            .to_dict()["loss"]
        )

        # Meta Data dict
        meta_data_dict = validation_df[["image_id", "meta_data"]].values
        meta_data_dict = dict(zip(meta_data_dict[:, 0], meta_data_dict[:, 1]))
        # Logits dict
        logits_dict = validation_df[["image_id", "logits"]].values
        logits_dict = dict(zip(logits_dict[:, 0], logits_dict[:, 1]))

        data["ground_truth"] = ground_truth_dict
        data["meta_data"] = meta_data_dict
        data["logits"] = logits_dict
        data["loss"] = loss_dict

        return data


class YOLOConverter:
    def __init__(self, annotations=None, successful_batch_data=None, image_width=None, image_height=None):
        self.annotations = annotations
        self.successful_batch_data = successful_batch_data
        self.image_width = image_width
        self.image_height = image_height

    def is_annot_yolo_format(self):
        """
        Check if the given annotations follow YOLO format.
        """
        if isinstance(self.annotations, dict):
            for annot in self.annotations.values():
                if isinstance(annot, list) and len(annot) == 1 and isinstance(annot[0], str):
                    try:
                        int(annot[0])
                        return True
                    except ValueError:
                        return False
        return False

    def is_pred_yolo_format(self):
        """
        Check if the given predictions follow YOLO format.
        YOLO predictions typically contain class, confidence, and optionally loss.
        """
        if isinstance(self.successful_batch_data, list) and len(self.successful_batch_data) > 0:
            pred = self.successful_batch_data[0]
            return all(key in pred for key in ['image_id', 'class', 'confidence'])
        return False

    def convert_annotations(self):
        """
        Convert YOLO annotations to the custom format used in the codebase.
        """
        custom_annotations = {}
        for image_id, annotations in self.annotations.items():
            custom_annotations[image_id] = {
                "annotation": [{
                    'id': 0,
                    'label': int(annotations[0])
                }]
            }
        return custom_annotations

    def convert_annot_if_needed(self):
        """
        Convert to custom format if the annotations are in YOLO format.
        """
        if self.is_annot_yolo_format():
            print("Annotations are in YOLO format. Converting to custom format.")
            return self.convert_annotations()
        else:
            print("Annotations are already in custom format. No conversion needed.")
            return self.annotations

    def convert_predictions(self):
        """
        Convert predictions from YOLO format to the custom format.
        """
        custom_predictions = {}
        for pred in self.successful_batch_data:
            image_id = pred['image_id']
            class_id = pred['class']
            confidence = pred['confidence']
            loss = pred.get('loss', None)
            
             # Initialize logits with zeros for two categories
            logits = [0.0, 0.0]
            logits[class_id] = confidence  # Set confidence at the index of category_id
            logits[1 - class_id] = 1 - confidence  # Set the remaining confidence value
            
            custom_predictions[image_id] = {
                'image_id': image_id,
                'prediction_class': class_id,
                'confidence': confidence,
                'logits': logits,
                'loss': loss
            }
        return custom_predictions

    def convert_pred_if_needed(self):
        """
        Convert to custom format if the predictions are in YOLO format.
        """
        if self.is_pred_yolo_format():
            print("Predictions are in YOLO format. Converting to custom format.")
            return self.convert_predictions()
        else:
            print("Predictions are already in custom format. No conversion needed.")
            return self.successful_batch_data


import numpy as np
import pandas as pd

class ValidationUtils:
    def __init__(self, meta_pred_true):
        if type(meta_pred_true) == dict:
            self.meta_pred_true = pd.DataFrame(meta_pred_true).T
        else:
            self.meta_pred_true = meta_pred_true

    def filter_attribute_by_dict(self, target_attribute_dict=None):
        """
        Filters data by more than one attribute.
        """
        if bool(target_attribute_dict) != False:
            all_params = target_attribute_dict.keys()
            filtered_meta_pred_true = self.meta_pred_true.copy()
            for target_attribute in all_params:
                if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                    slider_min, slider_max = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute].between(
                            slider_min, slider_max
                        )
                    ]
                else:
                    target_value = target_attribute_dict[target_attribute]
                    filtered_meta_pred_true = filtered_meta_pred_true[
                        self.meta_pred_true[target_attribute] == target_value
                    ]
            return filtered_meta_pred_true
        else:
            return self.meta_pred_true

    @staticmethod
    def filter_validation_collection_by_meta(validation_collection_data, meta_filter):
        filtered_validation_collection_data = []
        allowed_meta_keys = list(meta_filter.keys())
        for doc in validation_collection_data:
            add = ValidationUtils.is_meta_in_range(
                meta_data=doc.meta_data, meta_filter=meta_filter
            )
            if add:
                for meta_name in list(
                    set(doc.meta_data.keys()) - set(allowed_meta_keys)
                ):
                    doc.meta_data.pop(meta_name)
                filtered_validation_collection_data.append(doc.dict())
        return filtered_validation_collection_data

    @staticmethod
    def is_meta_in_range(meta_data, meta_filter):
        """
        Returns boolean value if given the meta value(meta filter) is in the meta data.
        """
        add = True
        for meta_key in meta_filter:
            if type(meta_filter[meta_key]) == str:
                if meta_data[meta_key] != meta_filter[meta_key]:
                    add = False
            elif type(meta_filter[meta_key]) == list:
                if meta_filter[meta_key][0] == "str":
                    for meta_key in meta_filter:
                        add_multi = []
                        if type(meta_filter[meta_key]) == str:
                            if meta_data[meta_key] != meta_filter[meta_key]:
                                add_multi.append(False)
                    add = any(add_multi)
                elif (
                    not min(meta_filter[meta_key])
                    <= meta_data[meta_key]
                    <= max(meta_filter[meta_key])
                ):
                    add = False
        return add

    def filter_attribute(self, target_attribute_dict):
        """
        Filters data by single attribute.
        """
        target_attribute = list(target_attribute_dict.keys())[0]
        target_value = target_attribute_dict[target_attribute]
        if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
            slider_min, slider_max = target_attribute_dict[target_attribute]
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute].between(slider_min, slider_max)
            ]
        else:
            filtered_meta_pred_true = self.meta_pred_true[
                self.meta_pred_true[target_attribute] == target_value
            ]
        return filtered_meta_pred_true

    def multifilter_attribute(self, target_attributes_dict):
        """
        Filters data by more than one attribute.
        """
        all_params = target_attributes_dict.keys()
        filtered_meta_pred_true = self.meta_pred_true.copy()
        for target_attribute in all_params:
            if self.is_list_numeric(self.meta_pred_true[target_attribute].values):
                slider_min, slider_max = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute].between(
                        slider_min, slider_max
                    )
                ]
            else:
                target_value = target_attributes_dict[target_attribute]
                filtered_meta_pred_true = filtered_meta_pred_true[
                    self.meta_pred_true[target_attribute] == target_value
                ]
        return filtered_meta_pred_true

    # Typecheckers
    @staticmethod
    def is_list_numeric(x_list):
        return all(
            [
                isinstance(
                    i,
                    (
                        int,
                        float,
                        np.int,
                        np.int8,
                        np.int16,
                        np.int32,
                        np.float,
                        np.float16,
                        np.float32,
                        np.float64,
                    ),
                )
                for i in x_list
            ]
        )

    @staticmethod
    def polygon_to_mask(poly, shape, max_value=1):
        import skimage

        if type(shape) == int:
            img = np.zeros((shape, shape, 1), "uint8")
        else:
            shape = int(shape[0]), int(shape[1]), 1
            img = np.zeros(shape)
        xs = [xy["x"] for xy in poly]
        ys = [xy["y"] for xy in poly]
        # fill polygon

        rr, cc = skimage.draw.polygon(xs, ys, img.shape)
        img[rr, cc] = max_value
        return img

    @staticmethod
    def rle_to_mask(mask_rle: str, shape, label=1):
        """
        mask_rle: run-length as string formatted (start length)
        shape: (height,width) of array to return
        Returns numpy array, 1 - mask, 0 - background

        """
        s = mask_rle.split()
        starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
        starts -= 1
        ends = starts + lengths
        img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
        for lo, hi in zip(starts, ends):
            img[lo:hi] = label
        return img.reshape(shape)  # Needed to align to RLE direction

    @staticmethod
    def mask_to_rle(image):
        """
        img: numpy array, 1 - mask, 0 - background
        Returns run length as string formatted
        Assumas image shape is (n,n) or (n,n,1)
        """
        image_shape = image.shape
        pixels = image.flatten()
        pixels = np.concatenate([[0], pixels, [0]])
        runs = np.where(pixels[1:] != pixels[:-1])[0] + 1
        runs[1::2] -= runs[::2]
        rle_string = " ".join(str(x) for x in runs)

        return rle_string, image_shape

    @staticmethod
    def polygon_to_rle(polygon: list, shape):
        import cv2
        from pycocotools.mask import encode
        
        mask = np.zeros(shape, dtype=np.uint8)
        polygon = np.asarray(polygon)
        polygon = polygon.reshape(-1, 2)
        cv2.fillPoly(mask, [polygon], 1)

        rle = encode(np.asfortranarray(mask))  # Encoding the mask into RLE
        rle["counts"] = rle["counts"].decode("utf-8")
        return rle

    @staticmethod
    def calculate_iou(gt_mask, pred_mask, threshold=0.5):
        """
        Calculate the Intersection over Union (IoU) of two bounding boxes.

        Parameters
        ----------
        gt_box : dict
            Keys: {'x1', 'x2', 'y1', 'y2'}
            The (x1, y1) position is at the top left corner,
            the (x2, y2) position is at the bottom right corner
        pred_box : dict
            Keys: {'x1', 'x2', 'y1', 'y2'}
            The (x, y) position is at the top left corner,
            the (x2, y2) position is at the bottom right corner

        Returns
        -------
        float
            in [0, 1]
        """
        # Don't even convert polygons to mask if there's no intersection
        gt_xs, gt_ys = [i["x"] for i in gt_mask], [i["y"] for i in gt_mask]
        min_gt_x, max_gt_x = min(gt_xs), max(gt_xs)
        min_gt_y, max_gt_y = min(gt_ys), max(gt_ys)

        pred_xs, pred_ys = [i["x"] for i in pred_mask], [i["y"] for i in pred_mask]
        min_pred_x, max_pred_x = min(pred_xs), max(pred_xs)
        min_pred_y, max_pred_y = min(pred_ys), max(pred_ys)

        shape = max(max_pred_x, max_gt_x, max_gt_y, max_pred_y)

        is_intersect = (max(min_gt_x, min_pred_x) < min(max_gt_x, max_pred_x)) and (
            max(min_gt_y, min_pred_y) < min(max_gt_y, max_pred_y)
        )
        if not is_intersect:
            iou = 0
            return iou

        mask1 = ValidationUtils.polygon_to_mask(gt_mask, shape=shape)
        mask2 = ValidationUtils.polygon_to_mask(pred_mask, shape=shape)

        mask1_area = np.count_nonzero(mask1 == 1)
        mask2_area = np.count_nonzero(mask2 == 1)
        intersection = np.count_nonzero(np.logical_and(mask1 == 1, mask2 == 1))
        iou = intersection / (mask1_area + mask2_area - intersection)
        return iou


class Statistics:
    @staticmethod
    def calculate_confidence_interval(metric, len_, z_value=1.96):
        metric = np.abs(metric)
        ci_length = z_value * np.sqrt((metric * (1 - metric)) / len_)
        ci_lower = metric - ci_length
        ci_upper = metric + ci_length
        return (ci_lower, ci_upper)

    @staticmethod
    def calculate_histogram(array_, min_, max_, n_bins):
        array = np.array(array_)
        bin_spaces = np.linspace(min_, max_, n_bins + 1)
        histogram_list = list()

        for i in range(len(bin_spaces) - 1):
            bin_min = bin_spaces[i]
            bin_max = bin_spaces[i + 1]
            histogram_list.append(
                {
                    "category": f"{bin_min.round(2)}",
                    "value": np.sum([(bin_min < array_) & (array_ <= bin_max)]).item(),
                }
            )

        return histogram_list


import time
import pandas as pd
import numpy as np

from .plots.plot_driver import SemanticSegmentationPlotDriver

class ValidationCreation:
    def __init__(self, batch_job_id, filter_field="image_url", generate_metrics=True):
        self.batch_job_id = batch_job_id
        self.filter_field = filter_field
        self.generate_metrics = generate_metrics

    def create_validation_collection_data(self, successful_batch_data, annotation_data, format=None):
        
        validation_collection_data = []
        
        for image_id in successful_batch_data:
            batch_item = successful_batch_data[image_id]
            annotation_item = annotation_data[image_id]
            image_information_dict = {}
            image_information_dict["batch_job_id"] = self.batch_job_id
            image_information_dict["image_id"] = batch_item["image_id"]
            image_information_dict["shape"] = [
                batch_item["shape"][0],
                batch_item["shape"][1],
            ]
            image_information_dict["meta_data"] = {},  # Placeholder for meta data if needed
            image_information_dict["ground_truth"] = annotation_item["annotation"]
            image_information_dict["objects"] = batch_item["masks"]
            image_information_dict["created_timestamp"] = time.time()
            validation_collection_data.append(image_information_dict)
            
        return validation_collection_data
    
    def load(self, validation_collection_data, class_mappings, study_list=None):

        generate_metrics = True
        
        #study_list = self.validation_cruds.get_study_list_by_batch_job_id(batch_job_id)
        #self.study_list = study_list
        self.study_list = []

        plotting_data = self._load_plotting_data(
            validation_collection_data=validation_collection_data,
            generate_metrics=generate_metrics,
            study_list=study_list,
        )

        # Create per image variables
        ground_truth_dict = plotting_data["per_image"]["ground_truth"]
        prediction_dict = plotting_data["per_image"]["prediction"]

        meta_data_dict = None

        try:
            meta_data_dict = plotting_data["per_image"]["meta_data"]
        except:
            print(
                "Meta Data not found.",
                )

        loss_dict = None

        try:
            loss_dict = plotting_data["per_image"]["loss"]
        except:
            print(
                "Loss not found.",
            )

        self.plot_driver = SemanticSegmentationPlotDriver(
            class_mappings=class_mappings,
            ground_truth_dict=ground_truth_dict,
            prediction_dict=prediction_dict,
            meta_data_dict=meta_data_dict,
            loss_dict=loss_dict,
            batch_job_id=self.batch_job_id,
            study_list=study_list,
        )
        
        overall_metrics = self.plot_driver._calling_all_plots()
        
        return overall_metrics
        
        
    def _load_plotting_data(
        self, validation_collection_data=None, generate_metrics=True, study_list=None,
    ):
        """
        Creates data for PlotValidation class(which gives payloads). Similar function in another class will be implemented for ModelValidation.

        :RaiseException: raises error if setup is not initialized.
        :return:  ground_truth_dict, logits_dict, meta_data_dict,loss_dict,img_source_dict
        """
        plotting_data = dict()
        plotting_data["per_image"] = self._craft_per_image_plotting_data(
            validation_collection_data, generate_metrics=generate_metrics, study_list=study_list
        )

        return plotting_data

    def _craft_per_image_plotting_data(
        self, validation_collection_data, generate_metrics, study_list=None
    ):
        """
        Creates data for PlotValidation class(which gives payloads). Similar function in another class will be implemented for ModelValidation.

        :RaiseException: raises error if setup is not initialized.
        :return:  ground_truth_dict, logits_dict, meta_data_dict,loss_dict,img_source_dict
        """
        data = dict()
        validation_df = pd.DataFrame(validation_collection_data)
        # Ground truth dict

        gt_dict = validation_df[["image_id", "ground_truth", "shape"]].values
        ground_truth_dict = {}
        for image_id, ground_truth_list, shape_list in gt_dict:
            shape = shape_list
            rle_dict = {"rles": []}
            if len(ground_truth_list) > 1:
                for item in ground_truth_list:
                    label = item["label"]
                    rle = {
                        "rle": item["mask"]["mask"], 
                        "shape": shape, 
                        "class": label
                    }
                    rle_dict["rles"].append(rle)
                ground_truth_dict[image_id] = rle_dict
                
            else:
                ground_truth = ground_truth_list[0]
                label = ground_truth["label"]
                rles_str = ground_truth["mask"]["mask"]
                rles = {
                    "rles": [{
                        "rle": rles_str, 
                        "shape": shape, 
                        "class": label
                    }]
                }
                ground_truth_dict[image_id] = rles

        # Prediction dict
        pred_dict = validation_df[["image_id", "objects", "shape"]].values
        prediction_dict = {}
        for image_id, objects, shape in pred_dict:
            rles = objects["rles"]
            for rle_dict in rles:
                rle_dict["shape"] = shape
            prediction_dict[image_id] = objects
        
        # Loss dict
        if generate_metrics:
            try:
                loss_dict = (
                    validation_df[["image_id", "loss"]]
                    .set_index("image_id")
                    .to_dict()["loss"]
                )
            except:
                pass

        # Meta Data dict
        try:
            meta_data_dict = validation_df[["image_id", "meta_data"]].values
            meta_data_dict = dict(zip(meta_data_dict[:, 0], meta_data_dict[:, 1]))
        except:
            pass
        # Image Sources dict

        data["ground_truth"] = ground_truth_dict
        data["prediction"] = prediction_dict
        
        try:
            data["meta_data"] = meta_data_dict
        except:
            pass
        if generate_metrics:
            try:
                data["loss"] = loss_dict
            except:
                pass
        return data


import json
import uuid
from collections import defaultdict
import pycocotools.mask as mask_utils

from gesund_val_library.utils.validation_data_utils import ValidationUtils

class COCOConverter:
    def __init__(self, problem_type, annotations=None, successful_batch_data=None):
        self.problem_type = problem_type
        if annotations:
            self.annotations = annotations
        if successful_batch_data:
            self.successful_batch_data = successful_batch_data

    def is_annot_coco_format(self):
        """
        Check if the given annotations follow COCO format.
        COCO format contains "images", "annotations", and "categories" keys.
        """
        return all(key in self.annotations for key in ['images', 'annotations', 'categories'])
    
    def is_pred_coco_format(self):
        """
        Check if the given predictions follow COCO format.
        COCO format for predictions typically contains "image_id", "category_id", and "score" keys.
        The "loss" key is optional.
        """
        if isinstance(self.successful_batch_data, list) and len(self.successful_batch_data) > 0:
            # Mandatory keys
            required_keys = {'image_id', 'category_id', 'score'}
            # Check if the first item contains all the required keys
            return all(key in self.successful_batch_data[0] for key in required_keys)
        return False

    def convert_annotations(self):
        if self.problem_type == 'classification':
            return self.convert_classification_annotations()
        elif self.problem_type == 'semantic_segmentation':
            return self.convert_semantic_segmentation_annotations()
        elif self.problem_type == 'object_detection':
            return self.convert_object_detection_annotations()
        elif self.problem_type == 'instance_segmentation':
            return self.convert_instance_segmentation_annotations()
        else:
            raise ValueError("Unsupported problem type.")

    def convert_predictions(self):
        if self.problem_type == 'classification':
            return self.convert_classification_predictions()
        elif self.problem_type == 'semantic_segmentation':
            return self.convert_semantic_segmentation_predictions()
        elif self.problem_type == 'object_detection':
            return self.convert_object_detection_predictions()
        elif self.problem_type == 'instance_segmentation':
            return self.convert_instance_segmentation_predictions()
        else:
            raise ValueError("Unsupported problem type.")

    def convert_classification_annotations(self):
        custom_annotations = {}
        for image in self.annotations['images']:
            image_id = image['id']
            custom_annotations[image_id] = {"annotation": []}

        for annotation in self.annotations['annotations']:
            image_id = annotation['image_id']
            custom_annotations[image_id]['annotation'].append({
                'id': annotation['id'],
                'label': annotation['category_id']
            })

        return custom_annotations

    def convert_semantic_segmentation_annotations(self):
        # Initialize a dictionary to hold final annotations
        custom_annotations = {}

        # Group annotations by image_id
        grouped_annotations = defaultdict(lambda: {
            "image_id": None,
            "annotation": []
        })

        for ann in self.annotations["annotations"]:
            image_id = ann["image_id"]
            category_id = ann["category_id"]
            rle = ann["segmentation"]
            size = rle["size"]

            # Convert RLE to mask string
            rle_mask = mask_utils.decode(rle)
            rle_, shape = ValidationUtils.mask_to_rle(rle_mask)

            # Populate the grouped_annotations structure
            if grouped_annotations[image_id]["image_id"] is None:
                grouped_annotations[image_id]["image_id"] = image_id

            # Create an annotation entry
            annotation_entry = {
                "image_id": image_id,
                "label": category_id,
                "type": "mask",
                "measurement_info": {
                    "objectName": "mask",
                    "measurement": "Segmentation"
                },
                "mask": {
                    "mask": rle_  # Use the RLE string generated
                },
                "shape": size,
                "window_level": None
            }

            # Append the annotation entry to the list for this image_id
            grouped_annotations[image_id]["annotation"].append(annotation_entry)

        # Transform grouped annotations into the desired format
        for image_id, data in grouped_annotations.items():
            custom_annotations[image_id] = {
                "image_id": data["image_id"],
                "annotation": data["annotation"]
            }

        return custom_annotations

            
    def convert_object_detection_annotations(self):
        custom_format = {}
        
        # Create a mapping of image IDs to file names and shapes
        images = {img['id']: {'file_name': img['file_name'], 'width': img['width'], 'height': img['height']} 
                for img in self.annotations['images']}
        annotations = self.annotations['annotations']
        
        for image_id, image_info in images.items():
            file_name = image_info['file_name']
            width = image_info['width']
            height = image_info['height']
            custom_annotations = []
            
            # Filter annotations for the current image
            for ann in annotations:
                if ann['image_id'] == image_id:
                    # Convert COCO bbox format (x, y, width, height) to custom points format
                    x1, y1, width_bbox, height_bbox = ann['bbox']
                    x2, y2 = x1 + width_bbox, y1 + height_bbox
                    points = [{"x": x1, "y": y1}, {"x": x2, "y": y2}]

                    # Append annotation to the list
                    custom_annotations.append({
                        "config_id": "uJ4FNIiXW4JyRCRRX_qzz",
                        "name": "default",
                        "id": str(uuid.uuid4()),
                        "label": ann['category_id'],  # You can adjust this mapping as needed
                        "points": points,
                        "type": "rect"
                    })
            
            # Add image and its annotations to the custom format
            custom_format[image_id] = {
                "image_id": image_id,
                "annotation": custom_annotations,
                "image_name": f"{file_name}",
                "shape": [height, width], 
                "last_updated_timestamp": 1727595047,
                "config_id": "uJ4FNIiXW4JyRCRRX_qzz"
            }
        
        return custom_format


    def convert_instance_segmentation_annotations(self):
        # Implement conversion logic for instance segmentation here
        pass

    def convert_classification_predictions(self):
        custom_predictions = {}
        for pred in self.successful_batch_data:
            image_id = pred['image_id']
            category_id = pred['category_id']
            confidence = pred['score']
            loss = pred.get('loss', None)

            logits = [0.0, 0.0]
            logits[category_id] = confidence
            logits[1 - category_id] = 1 - confidence

            custom_predictions[image_id] = {
                'image_id': image_id,
                'prediction_class': category_id,
                'confidence': confidence,
                'logits': logits,
                'loss': loss
            }

        return custom_predictions
    
    def convert_semantic_segmentation_predictions(self):
        custom_predictions = {}

        # Group by image_id
        grouped_predictions = defaultdict(lambda: {
            "image_id": None,
            "masks": {
                "rles": []
            },
            "shape": None,
            "status": 200
        })

        for pred in self.successful_batch_data:
            image_id = pred["image_id"]
            class_id = pred["category_id"]
            rle = pred["segmentation"]
            size = rle["size"]

            # Convert RLE to mask string
            rle_mask = mask_utils.decode([rle])
            rle_, shape = ValidationUtils.mask_to_rle(rle_mask)
            
            # Populate the grouped_annotations structure
            if grouped_predictions[image_id]["image_id"] is None:
                grouped_predictions[image_id]["image_id"] = image_id
                grouped_predictions[image_id]["shape"] = size
            
            # Create an RLE entry for the current annotation
            rle_entry = {
                "rle": rle_,  # Use the RLE string generated
                "class": class_id
            }
            
            # Append the RLE entry to the masks
            grouped_predictions[image_id]["masks"]["rles"].append(rle_entry)
            
            
        # Transform grouped annotations into the desired format
        for image_id, data in grouped_predictions.items():
            custom_predictions[image_id] = {
                "image_id": data["image_id"],
                "masks": data["masks"],
                "shape": data["shape"],
                "status": data["status"]
            }

        return custom_predictions

    def convert_object_detection_predictions(self):
        
        predictions_format = {}
    
        # Loop through each item in the custom format data
        for prediction in self.successful_batch_data:
            image_id = prediction['image_id']
            bbox = prediction['bbox']
            score = prediction['score']
            category_id = prediction['category_id']

            # Add to the predictions format
            if image_id not in predictions_format:
                predictions_format[image_id] = {
                    "objects": [],
                    "image_id": image_id,
                    "status": 200
                }
            
            # Append the object to the objects list
            predictions_format[image_id]['objects'].append({
                "box": {
                    "x1": round(bbox[0]),
                    "y1": round(bbox[1]),
                    "x2": round(bbox[0] + bbox[2]),  # Calculate x2 based on width
                    "y2": round(bbox[1] + bbox[3])   # Calculate y2 based on height
                },
                "confidence": round(score, 4),  # Use the provided score
                "prediction_class": category_id   # Use category_id directly
            })

        return predictions_format

    def convert_instance_segmentation_predictions(self):
        # Implement conversion logic for instance segmentation predictions here
        pass

    def convert_annot_if_needed(self):
        if self.is_annot_coco_format():
            print("Annotations are in COCO format. Converting to custom format.")
            return self.convert_annotations()
        else:
            print("Annotations are already in custom format. No conversion needed.")
            return self.annotations
        
    def convert_pred_if_needed(self):
        if self.is_pred_coco_format():
            print("Predictions are in COCO format. Converting to custom format.")
            return self.convert_predictions()
        else:
            print("Predictions are already in custom format. No conversion needed.")
            return self.successful_batch_data



import os
import warnings
from tqdm import tqdm
import pandas as pd

from .average_precision import PlotAveragePrecision
from .top_losses import PlotTopLosses
from .object_stats import PlotObjectStats
from .confidence_distribution import PlotConfidenceGraphs
from .dataset_stats import PlotDatasetStats

class ObjectDetectionPlotDriver:
    def __init__(
        self,
        coco_,
        class_mappings,
        ground_truth_dict,
        prediction_dict,
        batch_job_id,
        meta_data_dict=None,
        loss_dict=None,
    ):
        # Create validation variables
        self.true = ground_truth_dict
        self.pred = prediction_dict
        self.class_mappings = class_mappings
        self.meta = None
        self.batch_job_id = batch_job_id
        self.coco_ = coco_
        if meta_data_dict:
            self.meta = pd.DataFrame(meta_data_dict).T
        if loss_dict:
            self.loss = pd.DataFrame(loss_dict, index=[0])


        self.sample_size = len(self.true)
        self.class_order = list(range(len(class_mappings.keys())))
        # Check, add breakpoint

        # Import Classes
        self.plot_average_precision = PlotAveragePrecision(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
            coco_=coco_,
        )

        self.plot_object_stats = PlotObjectStats(
            coco_=coco_,
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
        )

        self.plot_confidence_graphs = PlotConfidenceGraphs(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
            coco_=coco_,
        )

        self.plot_dataset_stats = PlotDatasetStats(
            class_mappings=self.class_mappings,
            meta_data_dict=meta_data_dict,
        )

        self.plot_loss = PlotTopLosses(
            coco_=coco_,
            class_mappings=self.class_mappings,
            meta_dict=self.meta,
        )

    def plot_highlighted_overall_metrics(self):
        return self.plot_average_precision._plot_highlighted_overall_metrics(
            threshold=0.1
        )

    def plot_performance_by_iou_threshold(self, threshold=0.5, return_points=False):
        return self.plot_average_precision._plot_performance_by_iou_threshold(
            threshold, return_points
        )

    def plot_statistics_classbased_table(self, target_attribute_dict=None):
        return self.plot_average_precision._plot_statistics_classbased_table(
            threshold=0.1, target_attribute_dict=target_attribute_dict
        )

    def plot_object_counts(self, confidence=0, target_attribute_dict=None):
        return self.plot_object_stats._plot_object_counts(
            confidence=confidence, target_attribute_dict=target_attribute_dict
        )

    def plot_top_misses(self, top_k=100):
        return self.plot_loss._plot_top_misses(top_k=top_k)

    def plot_confidence_histogram_scatter_distribution(
        self, predicted_class=None, n_samples=300
    ):
        return (
            self.plot_confidence_graphs._plot_confidence_histogram_scatter_distribution(
                predicted_class, n_samples
            )
        )

    def plot_prediction_distribution(self, target_attribute_dict=None):
        return self.plot_object_stats._plot_prediction_distribution(
            target_attribute_dict
        )

    def plot_meta_distribution(
        self,
    ):
        return self.plot_dataset_stats._plot_meta_distributions()

    def plot_training_validation_comparison_classbased_table(self):
        return (
            self.plot_average_precision._plot_training_validation_comparison_classbased_table()
        )

    def main_metric(self):
        return self.plot_average_precision._main_metric(threshold=0.1)

    # Blind Spots
    def plot_blind_spot_metrics(self, target_attribute_dict=None):
        return self.plot_average_precision.blind_spot_metrics(
            target_attribute_dict, threshold=0.1
        )
        
    def _calling_all_plots(self):
        # Getting all methods that do not start with '_'
        methods = [
            method_name for method_name in dir(self)
            if callable(getattr(self, method_name)) and not method_name.startswith("_")
        ]
        results = {}

        for method_name in tqdm(methods, desc="Calling all plot functions"):
            method = getattr(self, method_name)
            try:
                # Suppress warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")

                    print(f"Calling method: {method_name}...")
                    # Attempt to call the method, handle cases where no arguments are required
                    result = method()
                    results[method_name] = result
            except TypeError as e:
                results[method_name] = f"Could not call {method_name}: {str(e)}"
            except Exception as e:
                results[method_name] = f"Error in {method_name}: {str(e)}"
        
        return results



import json
import bson
import os
from gesund_val_library.utils.io_utils import read_json, save_plot_metrics_as_json, format_metrics
from gesund_val_library.utils.yolo_converter import YOLOConverter
from gesund_val_library.utils.coco_converter import COCOConverter
from gesund_val_library.problem_type_factory import get_validation_creation

def run_metrics(args):
    """Run validation metrics based on the passed arguments."""
    try:
        successful_batch_data = read_json(args['predictions'])
        annotation_data = read_json(args['annotations_json_path'])
        class_mappings = read_json(args['class_mappings'])
    except (FileNotFoundError, json.JSONDecodeError) as e:
        print(f"Error loading input files: {e}")
        return None

    batch_job_id = str(bson.ObjectId())
    output_dir = os.path.join("outputs", batch_job_id)

    if args['format'] == 'coco':
        converter_annot = COCOConverter(annotations=annotation_data, problem_type=args['problem_type'])
        converter_pred = COCOConverter(successful_batch_data=successful_batch_data, problem_type=args['problem_type'])
        annotation_data = converter_annot.convert_annot_if_needed()
        successful_batch_data = converter_pred.convert_pred_if_needed()

    elif args['format'] == 'yolo':
        yolo_converter = YOLOConverter(annotations=annotation_data, successful_batch_data=successful_batch_data)
        annotation_data = yolo_converter.convert_annot_if_needed()
        successful_batch_data = yolo_converter.convert_pred_if_needed()

    elif args['format'] == 'gesund_custom_format':
        pass

    ValidationCreationClass = get_validation_creation(args['problem_type'])
    validation = ValidationCreationClass(batch_job_id)

    try:
        validation_data = validation.create_validation_collection_data(successful_batch_data, annotation_data, args['format'])
        metrics = validation.load(validation_data, class_mappings)
    except Exception as e:
        print(f"Error calculating metrics: {e}")
        return None

    #formatted_metrics = format_metrics(metrics)

    if args.get('write_results_to_json', False):
        save_plot_metrics_as_json(metrics, output_dir)

    return metrics



from setuptools import setup, find_packages

setup(
    name='gesund_val_library',
    version='0.3.3',
    author='Gesund AI',
    author_email='hammadbink@gmail.com',
    license="MIT",
    description='Gesund.ai package for running validation metrics for classification, semantic segmentation, instance segmentation, and object detection models.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/gesund-ai/gesund_val_library',
    packages=find_packages(),
    install_requires=[
        'bson',
        'jsonschema',
        'scikit-learn',
        'pandas',
        'pydicom',
        'nibabel',
        'opencv-python',
        'SimpleITK',
        'dictances==1.5.3',
        'miseval==1.2.2',
        'numpy==1.21.6',
        'numba==0.55.2',
        'tqdm',
        'pycocotools',
    ],
    dependency_links=[
        'git+https://github.com/HammadK44/cocoapi.git@Dev#egg=pycocotools&subdirectory=PythonAPI'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    include_package_data=True,
)



from gesund_val_library.metrics.classification.create_validation import ValidationCreation as ClassificationValidationCreation
from gesund_val_library.metrics.semantic_segmentation.create_validation import ValidationCreation as SegmentationValidationCreation
from gesund_val_library.metrics.object_detection.create_validation import ValidationCreation as ObjectDetectionValidationCreation

def get_validation_creation(problem_type):
    if problem_type == 'classification':
        return ClassificationValidationCreation
    elif problem_type == 'semantic_segmentation':
        return SegmentationValidationCreation
    #elif problem_type == 'instance_segmentation':
    #    return SegmentationValidationCreation
    elif problem_type == 'object_detection':
        return ObjectDetectionValidationCreation
    else:
        raise ValueError(f"Unknown problem type: {problem_type}")



import time
import datetime
import pandas as pd
import json
import numpy as np

from .plots.plot_driver import ObjectDetectionPlotDriver

class ValidationCreation:
    def __init__(self, batch_job_id, filter_field="image_url", generate_metrics=True):
        self.batch_job_id = batch_job_id
        self.filter_field = filter_field
        self.generate_metrics = generate_metrics

    def create_validation_collection_data(self, successful_batch_data, annotation_data, format):
        
        validation_collection_data = list()
        # To Do: Optimize this
        for image_id in successful_batch_data:
            batch_item = successful_batch_data[image_id]
            annotation_item = annotation_data[image_id]
            image_information_dict = {}
            image_information_dict["batch_job_id"] = self.batch_job_id
            image_information_dict["image_id"] = batch_item["image_id"]
            if format == 'coco':
                image_information_dict["shape"] = annotation_item["shape"]
            else:
                image_information_dict["shape"] = batch_item["shape"]
            image_information_dict["ground_truth"] = [
                {
                    "class": i["label"],
                    "box": {
                        "x1": i["points"][0]["x"],
                        "x2": i["points"][1]["x"],
                        "y1": i["points"][0]["y"],
                        "y2": i["points"][1]["y"],
                    },
                }
                for i in annotation_item["annotation"]
            ]
            image_information_dict["objects"] = batch_item["objects"]
            image_information_dict["created_timestamp"] = time.time()
            validation_collection_data.append(image_information_dict)
        return validation_collection_data
    
    
    def _load_plotting_data(
        self, validation_collection_data, class_mappings
    ):
        """
        Creates data for PlotValidation class(which gives payloads). Similar function in another class will be implemented for ModelValidation.

        :RaiseException: raises error if setup is not initialized.
        :return:  ground_truth_dict, logits_dict, meta_data_dict,loss_dict,img_source_dict
        """
        plotting_data = dict()
        plotting_data["per_image"] = self._craft_per_image_plotting_data(
            validation_collection_data
        )
        plotting_data["per_dataset"] = self._craft_per_dataset_plotting_data(
            validation_collection_data, class_mappings
        )
        return plotting_data

    def _craft_per_dataset_plotting_data(
        self, validation_collection_data, class_mappings
    ):

        per_dataset = dict()

        if validation_collection_data:
            per_dataset["prediction_coco"] = self._load_pred_coco(
                validation_collection_data
            )
            per_dataset["ground_truth_coco"] = self._load_annot_coco(
                validation_collection_data, class_mappings
            )

        return per_dataset

    def _craft_per_image_plotting_data(
        self, validation_collection_data
    ):
        """
        Creates data for PlotValidation class(which gives payloads). Similar function in another class will be implemented for ModelValidation.

        :RaiseException: raises error if setup is not initialized.
        :return:  ground_truth_dict, logits_dict, meta_data_dict,loss_dict,img_source_dict
        """
        data = dict()
        validation_df = pd.DataFrame(validation_collection_data)
        # Ground truth dict
        ground_truth_dict = validation_df[["image_id", "ground_truth"]].values
        ground_truth_dict = dict(zip(ground_truth_dict[:, 0], ground_truth_dict[:, 1]))
        # Prediction dict
        prediction_dict = validation_df[["image_id", "objects"]].values
        prediction_dict = dict(zip(prediction_dict[:, 0], prediction_dict[:, 1]))

        try:
            loss_dict = (
                validation_df[["image_id", "loss"]]
                .set_index("image_id")
                .to_dict()["loss"]
            )
        except:
            pass

        data["ground_truth"] = ground_truth_dict
        data["objects"] = prediction_dict

        try:
            data["loss"] = loss_dict
        except:
            pass
        return data

    def load(self, validation_collection_data, class_mappings, study_list=None):

        plotting_data = self._load_plotting_data(
            validation_collection_data=validation_collection_data, class_mappings=class_mappings
        )

        # Create per image variables
        ground_truth_dict = plotting_data["per_image"]["ground_truth"]
        prediction_dict = plotting_data["per_image"]["objects"]

        pred_coco = plotting_data["per_dataset"]["prediction_coco"]
        gt_coco = plotting_data["per_dataset"]["ground_truth_coco"]
        coco_ = [pred_coco, gt_coco]
        meta_data_dict = None

        try:
            meta_data_dict = plotting_data["per_image"]["meta_data"]
        except:
            print(
                "Meta Data not found.")

        loss_dict = None

        try:
            loss_dict = plotting_data["per_image"]["loss"]
        except:
            print(
                "Loss not found.")

        self.plot_driver = ObjectDetectionPlotDriver(
            coco_=coco_,
            class_mappings=class_mappings,
            ground_truth_dict=ground_truth_dict,
            prediction_dict=prediction_dict,
            meta_data_dict=meta_data_dict,
            loss_dict=loss_dict,
            batch_job_id=self.batch_job_id,
        )
        
        overall_metrics = self.plot_driver._calling_all_plots()

        # Remove mAP11@10 from overall_metrics
        if 'mAP11@10' in overall_metrics['main_metric']:
            del overall_metrics['main_metric']['mAP11@10']

        for key in overall_metrics['plot_blind_spot_metrics']:
            if 'AP11@10' in overall_metrics['plot_blind_spot_metrics'][key]:
                del overall_metrics['plot_blind_spot_metrics'][key]['AP11@10']

        if 'mAP11@10' in overall_metrics['plot_highlighted_overall_metrics']['data']:
            del overall_metrics['plot_highlighted_overall_metrics']['data']['mAP11@10']

        for key in overall_metrics['plot_statistics_classbased_table']['data']['Validation']:
            if 'AP11@10' in overall_metrics['plot_statistics_classbased_table']['data']['Validation'][key]:
                del overall_metrics['plot_statistics_classbased_table']['data']['Validation'][key]['AP11@10']
                
        return overall_metrics



    def _load_pred_coco(self, validation_collection_data=None):

        pred_dict_list = []
        predictions_list = validation_collection_data
        for pred in predictions_list:

            objects = pred["objects"]
            image_id = pred["image_id"]

            for object_ in objects:

                bbox = list(object_["box"].values())
                class_label = object_["prediction_class"]
                confidence = object_["confidence"]

                add_dict = {
                    "image_id": image_id,
                    "category_id": class_label,
                    "bbox": [
                        bbox[0],
                        bbox[1],
                        (bbox[2] - bbox[0]),
                        (bbox[3] - bbox[1]),
                    ],
                    "score": confidence,
                }
                pred_dict_list.append(add_dict)
        
        return pred_dict_list

    def _load_annot_coco(self, validation_collection_data=None, class_mappings=None):

        coco_dict = {
            "info": {
                "year": "2022",
                "version": "1.0",
                "description": "Exported from Gesund AI",
                "contributor": "Gesund AI",
                "url": "https://gesund.ai",
                "date_created": datetime.datetime.strftime(
                    datetime.datetime.now(), format="%Y-%m-%dT%H:%M:%S%z"
                ),
            },
            "licenses": [
                {
                    "url": "http://creativecommons.org/licenses/by-nc-sa/2.0/",
                    "id": 1,
                    "name": "NonLicensed Annotation",
                },
            ],
            "categories": [
                {"id": 0, "name": "cat", "supercategory": "animal"},
            ],
            "images": [
                {
                    "id": 0,
                    "license": 1,
                    "file_name": None,
                    "height": 480,
                    "width": 640,
                    "date_captured": None,
                },
            ],
            "annotations": [
                {
                    "id": 0,
                    "image_id": 0,
                    "category_id": 2,
                    "bbox": [260, 177, 231, 199],
                    "segmentation": None,
                    "area": 45969,
                    "iscrowd": 0,
                },
            ],
        }

        categories = []

        for id_, name in class_mappings.items():
            categories.append({"id": int(id_), "name": name, "supercategory": name})

        coco_images_list = []
        images_list = validation_collection_data
        for img in images_list:
            image_id = img["image_id"]
            file_name = f"{image_id}.jpg"
            image_dict = {
                "id": image_id,
                "license": 1,
                "file_name": file_name,
                "height": None,
                "width": None,
                "date_captured": None,
            }
            coco_images_list.append(image_dict)

        coco_annotations_list = []
        annotations_list = validation_collection_data
        obj_id = 0
        for ant in annotations_list:
            image_id = ant["image_id"]

            for p in ant["ground_truth"]:

                category = p["class"]

                ant_dict = {
                    "id": obj_id,
                    "image_id": image_id,
                    "category_id": category,
                    "bbox": [
                        p["box"]["x1"],
                        p["box"]["y1"],
                        (p["box"]["x2"] - p["box"]["x1"]),
                        (p["box"]["y2"] - p["box"]["y1"]),
                    ],
                    "segmentation": None,
                    "area": (p["box"]["x2"] - p["box"]["x1"])
                    * (p["box"]["y2"] - p["box"]["y1"]),
                    "iscrowd": 0,
                }
                obj_id += 1
                coco_annotations_list.append(ant_dict)

        coco_dict["categories"] = categories
        coco_dict["images"] = coco_images_list
        coco_dict["annotations"] = coco_annotations_list
        annot_dict = coco_dict

        return annot_dict



import argparse
import json
import bson
import os
from gesund_val_library.utils.io_utils import read_json, save_plot_metrics_as_json, format_metrics
from gesund_val_library.utils.yolo_converter import YOLOConverter
from gesund_val_library.utils.coco_converter import COCOConverter
from gesund_val_library.scripts.problem_type_factory import get_validation_creation

def main():
    parser = argparse.ArgumentParser(description='Run validation metrics calculation.')
    parser.add_argument('--annotations', type=str, required=True, help='Path to the JSON file with annotations.')
    parser.add_argument('--predictions', type=str, required=True, help='Path to the JSON file with predictions.')
    parser.add_argument('--class_mappings', type=str, required=True, help='Path to the JSON file with class mappings.')
    parser.add_argument('--problem_type', type=str, required=True,
                        choices=['classification', 'semantic_segmentation', 'instance_segmentation', 'object_detection'],
                        help='The type of problem.')
    parser.add_argument('--format', type=str, required=True, choices=['coco', 'yolo', 'gesund_custom_format'], 
                        help='Format of the input data (COCO, YOLO, Gesund_Custom_Format).')

    args = parser.parse_args()

    try:
        successful_batch_data = read_json(args.predictions)
        annotation_data = read_json(args.annotations)
        class_mappings = read_json(args.class_mappings)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        print(f"Error loading input files: {e}")
        return

    batch_job_id = str(bson.ObjectId())
    output_dir = os.path.join("outputs", batch_job_id)

    if args.format == 'coco':
        converter_annot = COCOConverter(annotations=annotation_data, problem_type=args.problem_type)
        converter_pred = COCOConverter(successful_batch_data=successful_batch_data, problem_type=args.problem_type)
        annotation_data = converter_annot.convert_annot_if_needed()
        successful_batch_data = converter_pred.convert_pred_if_needed()

    elif args.format == 'yolo':
        yolo_converter = YOLOConverter(annotations=annotation_data, successful_batch_data=successful_batch_data)
        annotation_data = yolo_converter.convert_annot_if_needed()
        successful_batch_data = yolo_converter.convert_pred_if_needed()

    elif args.format == 'gesund_custom_format':
        pass

    ValidationCreationClass = get_validation_creation(args.problem_type)
    validation = ValidationCreationClass(batch_job_id)

    try:
        validation_data = validation.create_validation_collection_data(successful_batch_data, annotation_data, args.format)
        metrics = validation.load(validation_data, class_mappings)
    except Exception as e:
        print(f"Error calculating metrics: {e}")
        return

    format_metrics(metrics)
    save_plot_metrics_as_json(metrics, output_dir)

if __name__ == "__main__":
    main()